@@ -1,5 +1,5 @@
 /*
- * Copyright 2011 - 2015
+ * Copyright 2011 - 2021
  * Andr\xe9 Malo or his licensors, as applicable
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -18,10 +18,9 @@
 #include "cext.h"
 EXT_INIT_FUNC;
 
-#ifdef EXT3
-typedef Py_UNICODE rchar;
-#else
 typedef unsigned char rchar;
+#ifdef U
+#undef U
 #endif
 #define U(c) ((rchar)(c))
 
@@ -161,6 +160,11 @@ static const rchar pattern_letter[] = {
     U('E'), U('T'), U('T'), U('E'), U('R')
 };
 
+static const rchar pattern_data[] = {
+    U('d'), U('a'), U('t'), U('a'), U(':'),
+    U('D'), U('A'), U('T'), U('A'), U(':')
+};
+
 static const rchar pattern_macie5_init[] = {
     U('/'), U('*'), U('\\'), U('*'), U('/')
 };
@@ -399,14 +403,44 @@ copy_uri_string(const rchar **source_, rchar **target_, rcssmin_ctx_t *ctx)
 {
     const rchar *source = *source_;
     rchar *target = *target_;
+    int is_spacy_data = 0;
     rchar c, quote = source[-1];
 
     *target++ = quote;
     *target_ = target;
 
+    if (!IMATCH(data, &source, &target, ctx)) {
+        --source;
+    }
+    else {
+        while (source < ctx->sentinel && target < ctx->tsentinel) {
+            c = *source++;
+            if (c <= U(' ') || c == U('\\') || c == U('"') || c == U('\'')) {
+                /* no match */
+                --source;
+                break;
+            }
+            *target++ = c;
+
+            if (c == U(',')) {
+                /* no need to check for boundaries below. The initial imatch
+                 * will block us from crossing them */
+                if (!(   target[-2] == U('4')
+                      && target[-3] == U('6')
+                      && (target[-4] == U('e') || target[-4] == U('E'))
+                      && (target[-5] == U('s') || target[-5] == U('S'))
+                      && (target[-6] == U('a') || target[-6] == U('A'))
+                      && (target[-7] == U('b') || target[-7] == U('B'))
+                      && target[-8] == U(';')))
+                    is_spacy_data = 1;
+                break;
+            }
+        }
+    }
+
     while (source < ctx->sentinel && target < ctx->tsentinel) {
         c = *source++;
-        if (RCSSMIN_IS_SPACE(c))
+        if (RCSSMIN_IS_SPACE(c) && !is_spacy_data)
             continue;
         *target++ = c;
         if (RCSSMIN_IS_STRING_DULL(c))
@@ -1033,12 +1067,15 @@ Minify CSS.\n\
 :Note: This is a hand crafted C implementation built on the regex\n\
        semantics.\n\
 \n\
-:Parameters:\n\
-  `style` : ``str``\n\
+Parameters:\n\
+  style (str):\n\
     CSS to minify\n\
 \n\
-:Return: Minified style\n\
-:Rtype: ``str``");
+  keep_bang_comments (bool):\n\
+    Keep comments starting with an exclamation mark? (``/*!...*/``)\n\
+\n\
+Returns:\n\
+  str: Minified style");
 
 static PyObject *
 rcssmin_cssmin(PyObject *self, PyObject *args, PyObject *kwds)
@@ -1049,15 +1086,15 @@ rcssmin_cssmin(PyObject *self, PyObject *args, PyObject *kwds)
     int keep_bang_comments;
 #ifdef EXT2
     int uni;
-#define UOBJ "O"
 #endif
 #ifdef EXT3
-#define UOBJ "U"
+    int bytes;
+    rchar *bytestyle;
 #endif
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, UOBJ "|O", kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|O", kwlist,
                                      &style, &keep_bang_comments_))
-        return NULL;
+        LCOV_EXCL_LINE_RETURN(NULL);
 
     if (!keep_bang_comments_)
         keep_bang_comments = 0;
@@ -1070,69 +1107,149 @@ rcssmin_cssmin(PyObject *self, PyObject *args, PyObject *kwds)
 #ifdef EXT2
     if (PyUnicode_Check(style)) {
         if (!(style = PyUnicode_AsUTF8String(style)))
-            return NULL;
+            LCOV_EXCL_LINE_RETURN(NULL);
         uni = 1;
     }
+    else if (!PyString_Check(style)) {
+        PyErr_SetString(PyExc_TypeError, "Unexpected type");
+        return NULL;
+    }
     else {
         if (!(style = PyObject_Str(style)))
-            return NULL;
+            LCOV_EXCL_LINE_RETURN(NULL);
         uni = 0;
     }
-#endif
-
-#ifdef EXT3
-    Py_INCREF(style);
-#define PyString_GET_SIZE PyUnicode_GET_SIZE
-#define PyString_AS_STRING PyUnicode_AS_UNICODE
-#define _PyString_Resize PyUnicode_Resize
-#define PyString_FromStringAndSize PyUnicode_FromUnicode
-#endif
 
+    /* Loop until the target buffer is big enough */
     rlength = slength = PyString_GET_SIZE(style);
+    while (1) {
+        if (!(result = PyString_FromStringAndSize(NULL, rlength))) {
+            LCOV_EXCL_START
 
-again:
-    if (!(result = PyString_FromStringAndSize(NULL, rlength))) {
-        Py_DECREF(style);
-        return NULL;
-    }
-    Py_BEGIN_ALLOW_THREADS
-    length = rcssmin((rchar *)PyString_AS_STRING(style),
-                     (rchar *)PyString_AS_STRING(result),
-                     slength, rlength, keep_bang_comments);
-    Py_END_ALLOW_THREADS
+            Py_DECREF(style);
+            return NULL;
 
-    if (length > rlength) {
-        Py_DECREF(result);
-        rlength = length;
-        goto again;
+            LCOV_EXCL_STOP
+        }
+        Py_BEGIN_ALLOW_THREADS
+        length = rcssmin((rchar *)PyString_AS_STRING(style),
+                         (rchar *)PyString_AS_STRING(result),
+                         slength, rlength, keep_bang_comments);
+        Py_END_ALLOW_THREADS
+
+        if (length > rlength) {
+            Py_DECREF(result);
+            rlength = length;
+            continue; /* Try again with a bigger buffer */
+        }
+        break;
     }
 
     Py_DECREF(style);
     if (length < 0) {
+        LCOV_EXCL_START
+
         Py_DECREF(result);
         return NULL;
-    }
-    if (length != rlength && _PyString_Resize(&result, length) == -1)
-        return NULL;
 
-#ifdef EXT2
+        LCOV_EXCL_STOP
+    }
+
+    if (length != rlength && _PyString_Resize(&result, length) == -1)
+        LCOV_EXCL_LINE_RETURN(NULL);
+
     if (uni) {
         style = PyUnicode_DecodeUTF8(PyString_AS_STRING(result),
                                      PyString_GET_SIZE(result), "strict");
         Py_DECREF(result);
-        if (!style)
-            return NULL;
-        result = style;
+        return style;
     }
-#endif
+
     return result;
+
+#else  /* EXT3 */
+
+    if (PyUnicode_Check(style)) {
+        bytes = 0;
+        style = PyUnicode_AsUTF8String(style);
+        bytestyle = (rchar *)PyBytes_AS_STRING(style);
+        slength = PyBytes_GET_SIZE(style);
+    }
+    else if (PyBytes_Check(style)) {
+        bytes = 1;
+        Py_INCREF(style);
+        bytestyle = (rchar *)PyBytes_AS_STRING(style);
+        slength = PyBytes_GET_SIZE(style);
+    }
+    else if (PyByteArray_Check(style)) {
+        bytes = 2;
+        Py_INCREF(style);
+        bytestyle = (rchar *)PyByteArray_AS_STRING(style);
+        slength = PyByteArray_GET_SIZE(style);
+    }
+    else {
+        PyErr_SetString(PyExc_TypeError, "Unexpected type");
+        return NULL;
+    }
+
+    /* Loop until the target buffer is big enough */
+    rlength = slength;
+    while (1) {
+        if (!(result = PyBytes_FromStringAndSize(NULL, rlength))) {
+            LCOV_EXCL_START
+
+            Py_DECREF(style);
+            return NULL;
+
+            LCOV_EXCL_STOP
+        }
+        Py_BEGIN_ALLOW_THREADS
+        length = rcssmin(bytestyle, (rchar *)PyBytes_AS_STRING(result),
+                         slength, rlength, keep_bang_comments);
+        Py_END_ALLOW_THREADS
+
+        if (length > rlength) {
+            Py_DECREF(result);
+            rlength = length;
+            continue; /* Try again with a bigger buffer */
+        }
+        break;
+    }
+
+    Py_DECREF(style);
+    if (length < 0) {
+        LCOV_EXCL_START
+
+        Py_DECREF(result);
+        return NULL;
+
+        LCOV_EXCL_STOP
+    }
+
+    if (!bytes) {
+        style = PyUnicode_DecodeUTF8(PyBytes_AS_STRING(result), length,
+                                     "strict");
+        Py_DECREF(result);
+        return style;
+    }
+    if (bytes == 1) {
+        if (length != slength) {
+            _PyBytes_Resize(&result, length);
+        }
+        return result;
+    }
+    /* bytes == 2: bytearray */
+    style = PyByteArray_FromStringAndSize(PyBytes_AS_STRING(result), length);
+    Py_DECREF(result);
+    return style;
+#endif
 }
 
 /* ------------------------ BEGIN MODULE DEFINITION ------------------------ */
 
 EXT_METHODS = {
     {"cssmin",
-        (PyCFunction)rcssmin_cssmin, METH_VARARGS | METH_KEYWORDS,
+        EXT_CFUNC(rcssmin_cssmin), METH_VARARGS | METH_KEYWORDS,
         rcssmin_cssmin__doc__},
 
     {NULL}  /* Sentinel */
@@ -1152,10 +1269,10 @@ EXT_INIT_FUNC {
 
     /* Create the module and populate stuff */
     if (!(m = EXT_CREATE(&EXT_DEFINE_VAR)))
-        EXT_INIT_ERROR(NULL);
+        EXT_INIT_ERROR(LCOV_EXCL_LINE(NULL));
 
     EXT_ADD_UNICODE(m, "__author__", "Andr\xe9 Malo", "latin-1");
-    EXT_ADD_STRING(m, "__docformat__", "restructuredtext en");
+    EXT_ADD_STRING(m, "__version__", STRINGIFY(EXT_VERSION));
 
     EXT_INIT_RETURN(m);
 }
